from functools import partialmethod
from typing import Any, Dict, Tuple


class _Meta(type):
    def __new__(mcs, name: str, bases: Tuple[Any], namespace: Dict[str, Any]) -> Any:
        if len(bases) == 0:
            return super().__new__(mcs, name, bases, namespace)

        cls_constructor = namespace.get("__init__")
        cls_params = getattr(cls_constructor, "__vedro__params__", None)
        if cls_params is None or len(cls_params) == 0:
            return super().__new__(mcs, name, bases, namespace)

        updated_name = "Template"
        updated_namespace = {**namespace, "__qualname__": updated_name}
        created = super().__new__(mcs, updated_name, bases, updated_namespace)

        cls_globals = getattr(cls_constructor, "__globals__")
        for idx, (args, kwargs) in enumerate(reversed(cls_params), start=1):
            cls_name = f"{name}_{idx}"
            cls_namespace = {
                **namespace,
                "__qualname__": cls_name,
                "__init__": partialmethod(cls_constructor, *args, **kwargs),
                "__vedro__template__": created,
            }
            cls_globals[cls_name] = type(cls_name, bases, cls_namespace)

        return created


class Scenario(metaclass=_Meta):
    pass



import inspect
from pathlib import Path
from typing import Any, List, Tuple, Type

from ..._scenario import Scenario
from .._scenario_finder import ScenarioFinder
from .._scenario_loader import ScenarioLoader
from .._virtual_scenario import VirtualScenario
from .._virtual_step import VirtualStep

__all__ = ("ScenarioDiscoverer",)


class ScenarioDiscoverer:
    def __init__(self, finder: ScenarioFinder, loader: ScenarioLoader) -> None:
        self._finder = finder
        self._loader = loader

    def _discover_steps(self, scenario: Type[Scenario]) -> List[VirtualStep]:
        steps = []
        for step in scenario.__dict__:
            if step.startswith("_"):
                continue
            method = getattr(scenario, step)
            if not inspect.isfunction(method):
                continue
            steps.append(VirtualStep(method))
        return steps

    async def discover(self, root: Path) -> List[VirtualScenario]:
        scenarios = []
        async for path in self._finder.find(root):
            loaded = await self._loader.load(path)
            for scn in loaded:
                steps = self._discover_steps(scn)
                scenarios.append(VirtualScenario(scn, steps))

        def cmp(scn: VirtualScenario) -> Tuple[Any, ...]:
            path = Path(scn.path)
            return (len(path.parts),) + tuple((len(x), x) for x in path.parts)
        scenarios.sort(key=cmp)

        return scenarios



from typing import Any, List, Union

from ._scenario_result import ScenarioResult

__all__ = ("Report",)


class Report:
    def __init__(self) -> None:
        self._results: List[ScenarioResult] = []
        self.started_at: Union[float, None] = None
        self.ended_at: Union[float, None] = None
        self.total: int = 0
        self.passed: int = 0
        self.failed: int = 0
        self.skipped: int = 0

    @property
    def elapsed(self) -> float:
        if (self.ended_at is None) or (self.started_at is None):
            return 0.0
        return self.ended_at - self.started_at

    def add_result(self, result: ScenarioResult) -> None:
        self._results.append(result)

        self.total += 1
        if result.is_passed():
            self.passed += 1
        elif result.is_failed():
            self.failed += 1
        elif result.is_skipped():
            self.skipped += 1

        if result.started_at:
            if self.started_at is None:
                self.started_at = result.started_at
            self.started_at = min(self.started_at, result.started_at)

        if result.ended_at:
            if self.ended_at is None:
                self.ended_at = result.ended_at
            self.ended_at = max(self.ended_at, result.ended_at)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)



from ._dispatcher import Dispatcher
from ._exc_info import ExcInfo
from ._lifecycle import Lifecycle
from ._plugin import Plugin
from ._report import Report
from ._runner import Runner
from ._scenario_discoverer import ScenarioDiscoverer
from ._scenario_finder import ScenarioFinder
from ._scenario_loader import ScenarioLoader
from ._scenario_result import ScenarioResult
from ._step_result import StepResult
from ._virtual_scenario import VirtualScenario
from ._virtual_step import VirtualStep

__all__ = ("Dispatcher", "ExcInfo", "Lifecycle", "Plugin", "Report", "Runner",
           "ScenarioDiscoverer", "ScenarioFinder", "ScenarioLoader",
           "ScenarioResult", "StepResult", "VirtualScenario", "VirtualStep",)



import os
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import List, Optional
from unittest.mock import Mock, call, sentinel

import pytest
from baby_steps import given, then, when
from rich.console import Console
from rich.style import Style

from vedro._core import (
    Dispatcher,
    ExcInfo,
    Report,
    ScenarioResult,
    StepResult,
    VirtualScenario,
    VirtualStep,
)
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
)
from vedro.plugins.director import RichReporter


@pytest.fixture()
def dispatcher():
    return Dispatcher()


@pytest.fixture()
def console_():
    return Mock(Console)


@pytest.fixture()
def reporter(console_):
    return RichReporter(lambda: console_)


def make_scenario_path(path: str = "", name: str = "scenario.py") -> Path:
    return Path(os.getcwd()) / "scenarios" / path / name


def make_scenario_result(*, scenario_path: Optional[Path] = None,
                         scenario_subject: Optional[str] = None,
                         step_results: Optional[List[StepResult]] = None) -> ScenarioResult:
    scenario_ = Mock(VirtualScenario)
    scenario_.path = scenario_path if scenario_path else make_scenario_path()
    scenario_.subject = scenario_subject

    scenario_result = ScenarioResult(scenario_)
    if step_results:
        for step_result in step_results:
            scenario_result.add_step_result(step_result)
    return scenario_result


def make_step_result(name: Optional[str] = None) -> StepResult:
    step_ = Mock(VirtualStep)
    step_.name = name if step_ else str(step_)
    return StepResult(step_)


def make_exc_info(value: Exception) -> ExcInfo:
    try:
        raise value
    except type(value):
        *_, traceback = sys.exc_info()
    return ExcInfo(type(value), value, traceback)


def make_parsed_args(*, verbose: int = 0,
                     tb_show_locals: bool = False,
                     tb_show_internal_calls: bool = False) -> Namespace:
    return Namespace(
        verbose=verbose,
        tb_show_internal_calls=tb_show_internal_calls,
        tb_show_locals=tb_show_locals,
    )


@pytest.mark.asyncio
async def test_rich_reporter_arg_parse_event(*, dispatcher: Dispatcher,
                                             reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        parser = ArgumentParser()
        event = ArgParseEvent(parser)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_arg_parsed_event(*, dispatcher: Dispatcher,
                                              reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        args = make_parsed_args()
        event = ArgParsedEvent(args)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_startup_event(*, dispatcher: Dispatcher,
                                           reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = StartupEvent([])

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out("Scenarios")
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_run_event(*, dispatcher: Dispatcher,
                                                reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        namespace = "<namespace>"
        scenario_result = make_scenario_result(scenario_path=make_scenario_path(namespace))
        event = ScenarioRunEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f"* {namespace}", style=Style.parse("bold"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_run_event_same_namespace(*, dispatcher: Dispatcher,
                                                               reporter: RichReporter,
                                                               console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        namespace = "<namespace>"
        scenario_result = make_scenario_result(scenario_path=make_scenario_path(namespace))
        event = ScenarioRunEvent(scenario_result)
        await dispatcher.fire(event)
        console_.reset_mock()

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_scenario_run_event_diff_namespace(*, dispatcher: Dispatcher,
                                                               reporter: RichReporter,
                                                               console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        namespace1 = "<namespace1>"
        scenario_result1 = make_scenario_result(scenario_path=make_scenario_path(namespace1))
        event1 = ScenarioRunEvent(scenario_result1)
        await dispatcher.fire(event1)
        console_.reset_mock()

        namespace2 = "<namespace2>"
        scenario_result2 = make_scenario_result(scenario_path=make_scenario_path(namespace2))
        event2 = ScenarioRunEvent(scenario_result2)

    with when:
        await dispatcher.fire(event2)

    with then:
        assert console_.mock_calls == [
            call.out(f"* {namespace2}", style=Style.parse("bold")),
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_skip_event(*, dispatcher: Dispatcher,
                                                 reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        scenario_result = make_scenario_result()
        event = ScenarioSkippedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_scenario_pass_event(*, dispatcher: Dispatcher,
                                                 reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        subject = "<subject>"
        scenario_result = make_scenario_result(scenario_subject=subject)
        event = ScenarioPassedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✔ {subject}", style=Style.parse("green"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_default_verbose(*, dispatcher: Dispatcher,
                                                                 reporter: RichReporter,
                                                                 console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        subject = "<subject>"
        scenario_result = make_scenario_result(scenario_subject=subject)
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✗ {subject}", style=Style.parse("red"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_verbose0(*, dispatcher: Dispatcher,
                                                          reporter: RichReporter,
                                                          console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(make_parsed_args(verbose=0))
        await dispatcher.fire(event)

        scenario_result = make_scenario_result()
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_verbose1(*, dispatcher: Dispatcher,
                                                          reporter: RichReporter,
                                                          console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(make_parsed_args(verbose=1))
        await dispatcher.fire(event)

        step_name_passed = "<passed step>"
        step_result_passed = make_step_result(step_name_passed).mark_passed()
        step_name_failed = "<failed step>"
        step_result_failed = make_step_result(step_name_failed).mark_failed()
        step_result = make_step_result()

        scenario_result = make_scenario_result(step_results=[
            step_result_passed,
            step_result_failed,
            step_result,
        ])
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red")),
            call.out(f"    ✔ {step_name_passed}", style=Style.parse("green")),
            call.out(f"    ✗ {step_name_failed}", style=Style.parse("red")),
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_verbose2(*, dispatcher: Dispatcher,
                                                          reporter: RichReporter,
                                                          console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(make_parsed_args(verbose=2))
        await dispatcher.fire(event)

        step_result_passed = make_step_result().mark_passed()
        step_result_failed = make_step_result().mark_failed()

        exc_info = make_exc_info(AssertionError())
        step_result_failed.set_exc_info(exc_info)
        step_result = make_step_result()

        scenario_result = make_scenario_result(step_results=[
            step_result_passed,
            step_result_failed,
            step_result,
        ])
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls[:3] == [
            call.out(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red")),
            call.out(f"    ✔ {step_result_passed.step_name}", style=Style.parse("green")),
            call.out(f"    ✗ {step_result_failed.step_name}", style=Style.parse("red")),
        ]
        name, args, kwargs = console_.mock_calls[3]
        assert "Traceback" in args[0]
        assert kwargs["style"] == Style.parse("yellow")


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_verbose3(*, dispatcher: Dispatcher,
                                                          reporter: RichReporter,
                                                          console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(make_parsed_args(verbose=3))
        await dispatcher.fire(event)

        step_result_passed = make_step_result("<passed step>").mark_passed()
        step_result_failed = make_step_result("<failed step>").mark_failed()
        exc_info = make_exc_info(AssertionError())
        step_result_failed.set_exc_info(exc_info)
        step_result = make_step_result()

        scenario_result = make_scenario_result(step_results=[
            step_result_passed,
            step_result_failed,
            step_result,
        ])
        scenario_result.set_scope({"key_int": 1, "key_str": "val"})
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls[:3] == [
            call.out(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red")),
            call.out(f"    ✔ {step_result_passed.step_name}", style=Style.parse("green")),
            call.out(f"    ✗ {step_result_failed.step_name}", style=Style.parse("red")),
        ]

        name, args, kwargs = console_.mock_calls[3]
        assert "Traceback" in args[0]
        assert kwargs["style"] == Style.parse("yellow")

        name, args, kwargs = console_.mock_calls[4]
        assert "Scope" in args[0]
        assert kwargs["style"] == Style(color="blue", bold=True)


@pytest.mark.asyncio
async def test_rich_reporter_scenario_fail_event_without_steps_verbose3(*, dispatcher: Dispatcher,
                                                                        reporter: RichReporter,
                                                                        console_: Mock):
    with given:
        reporter.subscribe(dispatcher)
        event = ArgParsedEvent(make_parsed_args(verbose=3))
        await dispatcher.fire(event)

        scenario_result = make_scenario_result()
        scenario_result.set_scope({})
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✗ {scenario_result.scenario_subject}", style=Style.parse("red")),
        ]


@pytest.mark.asyncio
async def test_rich_reporter_cleanup_event(*, dispatcher: Dispatcher):
    with given:
        console_ = Mock()
        reporter = RichReporter(lambda: console_)
        reporter.subscribe(dispatcher)

        report = Report()
        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(" "),
            call.out("# 0 scenarios, 0 passed, 0 failed, 0 skipped",
                     style=Style.parse("bold red"),
                     end=""),
            call.out(" (0.00s)", style=Style.parse("blue"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_success_cleanup_event(*, dispatcher: Dispatcher,
                                                   reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        report = Report()
        scenario_result1 = make_scenario_result().mark_passed()
        scenario_result1.set_started_at(1.0)
        scenario_result1.set_ended_at(2.0)
        report.add_result(scenario_result1)

        scenario_result2 = make_scenario_result().mark_passed()
        scenario_result2.set_started_at(2.0)
        scenario_result2.set_ended_at(3.0)
        report.add_result(scenario_result2)

        scenario_result3 = make_scenario_result().mark_skipped()
        report.add_result(scenario_result3)
        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(" "),
            call.out("# 3 scenarios, 2 passed, 0 failed, 1 skipped",
                     style=Style.parse("bold green"),
                     end=""),
            call.out(" (2.00s)", style=Style.parse("blue"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_failed_cleanup_event(*, dispatcher: Dispatcher,
                                                  reporter: RichReporter, console_: Mock):
    with given:
        reporter.subscribe(dispatcher)

        report = Report()
        scenario_result = make_scenario_result().mark_failed()
        scenario_result.set_started_at(3.145)
        scenario_result.set_ended_at(6.285)
        report.add_result(scenario_result)

        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(" "),
            call.out("# 1 scenarios, 0 passed, 1 failed, 0 skipped",
                     style=Style.parse("bold red"),
                     end=""),
            call.out(" (3.14s)", style=Style.parse("blue"))
        ]


def test_rich_reporter_format_scope_without_values(*, reporter: RichReporter):
    with given:
        scope = {}

    with when:
        res = list(reporter._format_scope(scope))

    with then:
        assert res == []


def test_rich_reporter_format_scope_with_values(*, reporter: RichReporter):
    with given:
        scope = {"key_int": 1, "key_str": "val"}

    with when:
        res = list(reporter._format_scope(scope))

    with then:
        assert res == [
            ("key_int", "1"),
            ("key_str", '"val"'),
        ]


def test_rich_reporter_format_scope_with_unserializable_value(*, reporter: RichReporter):
    with given:
        unserializable = sentinel
        scope = {"key_int": 1, "key_unserializable": unserializable}

    with when:
        res = list(reporter._format_scope(scope))

    with then:
        assert res == [
            ("key_int", "1"),
            ("key_unserializable", repr(unserializable)),
        ]



from types import TracebackType
from typing import Type

__all__ = ("ExcInfo",)


class ExcInfo:
    def __init__(self,
                 type_: Type[BaseException],
                 value: BaseException,
                 traceback: TracebackType) -> None:
        self.type = type_
        self.value = value
        self.traceback = traceback

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.type!r}, {self.value!r}, {self.traceback!r})"



from importlib.abc import Loader
from types import ModuleType
from typing import Any, Optional

from dessert import AssertionRewritingHook  # type: ignore

from ._scenario_file_loader import ScenarioFileLoader

__all__ = ("ScenarioAssertRewriterLoader",)

RewriterType = Any


class ScenarioAssertRewriterLoader(ScenarioFileLoader):
    def __init__(self, rewriter: Optional[RewriterType] = None) -> None:
        self._rewriter = rewriter or AssertionRewritingHook()

    def _exec_module(self, loader: Loader, module: ModuleType) -> None:
        self._rewriter.exec_module(module)



import re

from ..._core import Plugin

__all__ = ("Reporter",)


class Reporter(Plugin):
    @property
    def name(self) -> str:
        cls_name = self.__class__.__name__
        if cls_name.endswith(Reporter.__name__):
            cls_name = cls_name[:-len(Reporter.__name__)]
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls_name).lower()



import sys

from baby_steps import given, then, when

from vedro._core import ExcInfo


def test_exc_info():
    with given:
        try:
            raise AssertionError()
        except AssertionError:
            exc_type, exc_value, exc_tb = sys.exc_info()

    with when:
        exc_info = ExcInfo(exc_type, exc_value, exc_tb)

    with then:
        assert exc_info.type == exc_type
        assert exc_info.value == exc_value
        assert exc_info.traceback == exc_tb

        assert repr(exc_info) == f"ExcInfo(<class 'AssertionError'>, AssertionError(), {exc_tb!r})"



import sys

from pytest import raises

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when

from vedro import Scenario
from vedro._core import Dispatcher, Runner, VirtualScenario, VirtualStep
from vedro.events import (
    ExceptionRaisedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    StepFailedEvent,
    StepPassedEvent,
    StepRunEvent,
)


@pytest.fixture()
def dispatcher_():
    return AsyncMock(Dispatcher())


@pytest.fixture()
def runner(dispatcher_: Dispatcher):
    return Runner(dispatcher_)


@pytest.mark.asyncio
@pytest.mark.parametrize("method_mock_factory", (Mock, AsyncMock,))
async def test_runner_run_step_passed(method_mock_factory: Mock, *,
                                      runner: Runner, dispatcher_: Dispatcher):
    with given:
        scenario_ = Mock(Scenario, step=method_mock_factory(return_value=None))
        step = VirtualStep(scenario_.step)

    with when:
        step_result = await runner.run_step(step, scenario_)

    with then:
        assert scenario_.mock_calls == [call.step(scenario_)]

        assert step_result.is_passed() is True
        assert step_result.exc_info is None
        assert isinstance(step_result.started_at, float)
        assert isinstance(step_result.ended_at, float)

        assert dispatcher_.mock_calls == [
            call.fire(StepRunEvent(step_result)),
            call.fire(StepPassedEvent(step_result)),
        ]


@pytest.mark.asyncio
@pytest.mark.parametrize("method_mock_factory", (Mock, AsyncMock))
async def test_runner_run_step_failed(method_mock_factory: Mock, *,
                                      runner: Runner, dispatcher_: Dispatcher):
    with given:
        exception = AssertionError()
        scenario_ = Mock(Scenario, step=method_mock_factory(side_effect=exception))
        step = VirtualStep(scenario_.step)

    with when:
        step_result = await runner.run_step(step, scenario_)

    with given:
        assert scenario_.mock_calls == [call.step(scenario_)]

        assert step_result.is_failed() is True

        assert step_result.exc_info.value == exception
        assert isinstance(step_result.started_at, float)
        assert isinstance(step_result.ended_at, float)

        assert dispatcher_.mock_calls == [
            call.fire(StepRunEvent(step_result)),
            call.fire(ExceptionRaisedEvent(step_result.exc_info)),
            call.fire(StepFailedEvent(step_result)),
        ]


@pytest.mark.asyncio
@pytest.mark.parametrize("method_mock_factory", (Mock, AsyncMock))
async def test_runner_run_step_interrupted(*, method_mock_factory: Mock, dispatcher_: Dispatcher):
    with given:
        interrupt_exception = KeyboardInterrupt
        scenario_ = Mock(Scenario, step=method_mock_factory(side_effect=interrupt_exception))
        virtual_step = VirtualStep(scenario_.step)

        runner = Runner(dispatcher_, interrupt_exceptions=(interrupt_exception,))

    with when, raises(BaseException) as exception:
        await runner.run_step(virtual_step, scenario_)

    with given:
        assert exception.type is interrupt_exception
        assert scenario_.mock_calls == [call.step(scenario_)]


@pytest.mark.asyncio
async def test_runner_run_scenario_no_steps_passed(*, runner: Runner, dispatcher_: Dispatcher):
    with given:
        scenario_ = Mock(Scenario, step=Mock(return_value=None), __file__="/tmp/scenario.py")
        virtual_scenario = VirtualScenario(scenario_, [])

    with when:
        scenario_result = await runner.run_scenario(virtual_scenario)

    with then:
        assert scenario_result.is_passed() is True
        assert isinstance(scenario_result.started_at, float)
        assert isinstance(scenario_result.ended_at, float)

        assert dispatcher_.mock_calls == [
            call.fire(ScenarioRunEvent(scenario_result)),
            call.fire(ScenarioPassedEvent(scenario_result)),
        ]


@pytest.mark.asyncio
async def test_runner_run_scenario_single_step_passed(*, runner: Runner, dispatcher_: Dispatcher):
    with given:
        scenario_ = Mock(Scenario, step=Mock(return_value=None), __file__="/tmp/scenario.py")
        step = VirtualStep(scenario_.step)
        virtual_scenario = VirtualScenario(scenario_, [step])

    with when:
        scenario_result = await runner.run_scenario(virtual_scenario)

    with then:
        assert scenario_result.is_passed() is True
        assert isinstance(scenario_result.started_at, float)
        assert isinstance(scenario_result.ended_at, float)

        step_results = scenario_result.step_results
        assert dispatcher_.mock_calls == [
            call.fire(ScenarioRunEvent(scenario_result)),
            call.fire(StepRunEvent(step_results[0])),
            call.fire(StepPassedEvent(step_results[0])),
            call.fire(ScenarioPassedEvent(scenario_result)),
        ]


@pytest.mark.asyncio
async def test_runner_run_scenario_single_step_failed(*, runner: Runner, dispatcher_: Dispatcher):
    with given:
        exception = AssertionError()
        scenario_ = Mock(Scenario, step=Mock(side_effect=exception), __file__="/tmp/scenario.py")

        step = VirtualStep(scenario_.step)
        virtual_scenario = VirtualScenario(scenario_, [step])

    with when:
        scenario_result = await runner.run_scenario(virtual_scenario)

    with then:
        assert scenario_result.is_failed() is True
        assert isinstance(scenario_result.started_at, float)
        assert isinstance(scenario_result.ended_at, float)

        step_results = scenario_result.step_results
        assert dispatcher_.mock_calls == [
            call.fire(ScenarioRunEvent(scenario_result)),
            call.fire(StepRunEvent(step_results[0])),
            call.fire(ExceptionRaisedEvent(step_results[0].exc_info)),
            call.fire(StepFailedEvent(step_results[0])),
            call.fire(ScenarioFailedEvent(scenario_result)),
        ]


@pytest.mark.asyncio
async def test_runner_run_scenario_multiple_steps_passed(*, runner: Runner,
                                                         dispatcher_: Dispatcher):
    with given:
        scenario_ = Mock(Scenario, __file__="/tmp/scenario.py",
                         first_step=Mock(return_value=None),
                         second_step=Mock(return_value=None))

        first_step = VirtualStep(scenario_.first_step)
        second_step = VirtualStep(scenario_.second_step)
        virtual_scenario = VirtualScenario(scenario_, [first_step, second_step])

    with when:
        scenario_result = await runner.run_scenario(virtual_scenario)

    with then:
        assert scenario_result.is_passed() is True
        assert isinstance(scenario_result.started_at, float)
        assert isinstance(scenario_result.ended_at, float)

        first_step_result = scenario_result.step_results[0]
        second_step_result = scenario_result.step_results[1]
        assert dispatcher_.mock_calls == [
            call.fire(ScenarioRunEvent(scenario_result)),

            call.fire(StepRunEvent(first_step_result)),
            call.fire(StepPassedEvent(first_step_result)),

            call.fire(StepRunEvent(second_step_result)),
            call.fire(StepPassedEvent(second_step_result)),

            call.fire(ScenarioPassedEvent(scenario_result)),
        ]


@pytest.mark.asyncio
async def test_runner_run_scenario_multiple_steps_failed():
    with given:
        dispatcher = AsyncMock(Dispatcher)
        runner = Runner(dispatcher)

        exception = AssertionError()
        scenario_ = Mock(Scenario, __file__="/tmp/scenario.py",
                         first_step=Mock(return_value=None),
                         second_step=Mock(side_effect=exception),
                         third_step=Mock(return_value=None))

        first_step = VirtualStep(scenario_.first_step)
        second_step = VirtualStep(scenario_.second_step)
        third_step = VirtualStep(scenario_.third_step)
        scenario = VirtualScenario(scenario_, [first_step, second_step, third_step])

    with when:
        scenario_result = await runner.run_scenario(scenario)

    with then:
        assert scenario_result.is_failed() is True
        assert isinstance(scenario_result.started_at, float)
        assert isinstance(scenario_result.ended_at, float)

        first_step_result = scenario_result.step_results[0]
        second_step_result = scenario_result.step_results[1]
        assert dispatcher.mock_calls == [
            call.fire(ScenarioRunEvent(scenario_result)),

            call.fire(StepRunEvent(first_step_result)),
            call.fire(StepPassedEvent(first_step_result)),

            call.fire(StepRunEvent(second_step_result)),
            call.fire(ExceptionRaisedEvent(second_step_result.exc_info)),
            call.fire(StepFailedEvent(second_step_result)),

            call.fire(ScenarioFailedEvent(scenario_result)),
        ]


@pytest.mark.asyncio
async def test_runner_interrupted_scenario(*, dispatcher_: Dispatcher):
    with given:
        interrupt_exception = KeyboardInterrupt
        runner = Runner(dispatcher_, interrupt_exceptions=(interrupt_exception,))

        step_ = Mock(side_effect=interrupt_exception)
        scenario_ = Mock(Scenario, step=step_, __file__="/tmp/scenario.py")
        virtual_scenario = VirtualScenario(scenario_, [VirtualStep(step_)])

    with when, raises(BaseException) as exception:
        await runner.run_scenario(virtual_scenario)

    with then:
        assert exception.type is interrupt_exception



from ._director import Director
from ._reporter import Reporter
from .rich import RichReporter
from .silent import SilentReporter

__all__ = ("Director", "Reporter", "RichReporter", "SilentReporter",)



import os
from unittest.mock import Mock, patch

import pytest
from baby_steps import given, then, when
from rich.console import Console

from vedro.plugins.director import RichReporter
from vedro.plugins.director.rich.utils import get_terminal_size, make_console


@pytest.fixture()
def console_():
    return Mock(Console)


@pytest.fixture()
def reporter(console_):
    return RichReporter(lambda: console_)


def test_make_console():
    with when:
        res = make_console()

    with then:
        assert isinstance(res, Console)


def test_make_console_custom_options():
    with given:
        width, height = 1, 2

    with when:
        res = make_console(width=width, height=height)

    with then:
        assert isinstance(res, Console)
        assert res.size == (width, height)


def test_get_terminal_size():
    with given:
        width, height = 1, 2
        terminal_size = os.terminal_size((width, height))

    with when, patch("shutil.get_terminal_size", return_value=terminal_size):
        res = get_terminal_size()

    with then:
        assert res == terminal_size


def test_get_terminal_size_default_fallback():
    with given:
        width, height = 80, 24
        terminal_size = os.terminal_size((0, 0))

    with when, patch("shutil.get_terminal_size", return_value=terminal_size):
        res = get_terminal_size(width, height)

    with then:
        assert res == os.terminal_size((80, 24))


def test_get_terminal_size_custom_fallback():
    with given:
        width, height = 1, 2
        terminal_size = os.terminal_size((0, 0))

    with when, patch("shutil.get_terminal_size", return_value=terminal_size):
        res = get_terminal_size(width, height)

    with then:
        assert res == os.terminal_size((width, height))



from unittest.mock import Mock

from baby_steps import given, then, when

from vedro._core import Dispatcher
from vedro.plugins.director import SilentReporter


def test_silent_reporter():
    with given:
        dispatcher_ = Mock(Dispatcher())
        reporter = SilentReporter()

    with when:
        res = reporter.subscribe(dispatcher_)

    with then:
        assert res is None
        assert dispatcher_.mock_calls == []



from typing import List, Optional, Union

from ..._core import Dispatcher, Plugin
from ...events import ArgParseEvent
from ._reporter import Reporter

__all__ = ("Director",)


class Director(Plugin):
    def __init__(self, reporters: Optional[List[Reporter]] = None) -> None:
        self._reporters = {r.name: r for r in reporters} if reporters else {}
        self._default_reporter = reporters[0].name if reporters else ""
        self._dispatcher: Union[Dispatcher, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        self._dispatcher = dispatcher.listen(ArgParseEvent, self.on_arg_parse)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        if len(self._reporters) == 0:
            return
        event.arg_parser.add_argument("-r", "--reporters", nargs='*', choices=self._reporters,
                                      default=[self._default_reporter], help="Set reporter")
        args, *_ = event.arg_parser.parse_known_args()
        for reporter_name in args.reporters:
            reporter = self._reporters[reporter_name]
            assert self._dispatcher is not None
            reporter.subscribe(self._dispatcher)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import AsyncGenerator

__all__ = ("ScenarioFinder",)


class ScenarioFinder(ABC):
    @abstractmethod
    async def find(self, root: Path) -> AsyncGenerator[Path, None]:
        # https://github.com/python/mypy/issues/5070
        yield Path()



from pathlib import Path
from textwrap import dedent

import pytest

from vedro._core._scenario_loader import ScenarioAssertRewriterLoader


@pytest.mark.asyncio
async def test_scenario_assert_rewriter_loader(tmp_path: Path):
    path = tmp_path / "scenario.py"
    path.write_text(dedent('''
        import vedro
        class Scenario(vedro.Scenario):
            def then(self):
                assert 1 == 2
    '''))

    loader = ScenarioAssertRewriterLoader()
    scenarios = await loader.load(path)
    assert len(scenarios) == 1

    scenario = scenarios[0]()
    with pytest.raises(AssertionError) as exception:
        scenario.then()

    assert str(exception.value) == "assert 1 == 2"
